/* Reed 5.3 - An autoscrolling text pager
   Copyright (C)2000-2002 Joe Wreschnig <piman@sacredchao.net>

   next_bracket() Copyright (C)2000 Ben Zeigler, 2002 Joe Wreschnig
   getuser(), getgroup() Copyright (C)1985, 1990, 1993, 1998-2000
                                      Free Software Foundation, Inc.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/* Dar Williams rocks my world. */

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/dir.h>

#include <ctype.h>
#include <curses.h>
#include <dirent.h>
#include <errno.h>
#include <getopt.h>
#include <grp.h>
#include <limits.h>
#include <pwd.h>
#include <regex.h> 
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#include "filemode.h"

#define free_regexp(n) if (n) { regfree(n); free(n); } n = NULL;
#define alert() if (cue) beep();

#define VERSION "5.3"

/* These are return values for display_file_node */
#define QUIT	        0
#define NEXT_FILE	1
#define PREV_FILE	2
#define DELETE_FILE	3
#define LOAD_FILE       4
#define RELOAD_FILE     5
#define BOOKMARKS       6
#define HELP            7

/* This is some text I'm	putting here because I often use Reed's source code as a test for its
  own capabilities, so I need something in here that overflows a standard 80 column window, and then another line that overflows twice. That should be enough for most tests. Don't worry, there won't be control characters here.
                      					                     Foo
*/

WINDOW *text = NULL, *status = NULL;
char *bookmark_file = NULL;
int initial_delay = 500000, initial_jump = 1, cue = 1, initial_pause = 1;

static struct userid *user_alist;
static struct userid *group_alist;

typedef struct _FileNode FileNode;
struct _FileNode {
  int line, lines, *markers, paused, is_file;
  signed int jump;
  unsigned long delay;	/* Stored as microseconds/10 */
  char *filename, *message;
  FILE *file;
  regex_t *match;
  FileNode *n, *p;
};

struct userid {
  union {
    uid_t u;
    gid_t g;
  } id;
  char *name;
  struct userid *next;
};

void *xmalloc(size_t size) {
  void *value = malloc(size);
  if (value == NULL) {
    perror("reed");
    exit(1);
  }
  return value;
}

FileNode *new_file_node(void);
void delete_file_node(FileNode *node);
FileNode *next_file_node(FileNode *node);
FileNode *prev_file_node(FileNode *node);

FileNode *load_file(const char *filename);
void find_line_markers(FileNode *node);

void pause_file(FileNode *node);
void unpause_file(FileNode *node);
void toggle_pause(FileNode *node);

int seek_to_line_and_display(FileNode *node, int line);
void update_status_bar(FileNode *node);

char *input_string(const char *message);
float input_decimal(const char *message);
int input_integer(const char *message);

int get_bookmark(const char *name, FileNode *node);
void delete_bookmark(const char *name, FileNode *node);
void set_bookmark(const char *name, FileNode *node);
void clear_bookmarks(FileNode *node);
void set_char_bookmark(char c, FileNode *node);
int get_char_bookmark(char c, FileNode *node);

regex_t *make_regexp(const char *s);
int search_for(regex_t *match, FileNode *node, int up_down);
void new_search(FileNode *node, int up_down);
void next_bracket(const char *left, const char *right, FileNode *node,
                  int direction);

int set_active_file_node(FileNode *node);

FileNode *load_new_node(FileNode *old_node, const char *filename);
void parse_rc_file(void);

void initialize_windows(void);
void spawn_editor(FileNode *node);

char *getuser (uid_t uid);
char *getgroup (gid_t gid);

void print_help();
void print_version();

void parse_stdin(FileNode *n);

FileNode *new_file_node() {
  FileNode *node = xmalloc(sizeof(FileNode));
  node->lines = node->line = 0;
  node->paused = initial_pause;
  node->is_file = 1;
  node->delay = initial_delay;
  node->jump = initial_jump;
  node->n = node->p = NULL;
  node->file = NULL;
  node->filename = node->message = NULL;
  node->markers = NULL;
  node->match = NULL;
  return node;
}

void delete_file_node(FileNode *node) {
  if (node->n) node->n->p = node->p;
  if (node->p) node->p->n = node->n;
  if (node->match) regfree(node->match);
  if (node->file) fclose(node->file);
  free(node->match);
  free(node->filename);
  free(node->markers);
  free(node);
}

FileNode *next_file_node(FileNode *node) {
  if (node->n != NULL) return node->n;
  node->message = "You are at the last node.";
  return node;
}

FileNode *prev_file_node(FileNode *node) {
  if (node->p != NULL) return node->p;
  node->message = "You are at the first node.";
  return node;
}

int one(struct dirent *unused) { return 1; }

/* From the GNU fileutils package */
char *getuser (uid_t uid) {
  struct userid *tail;
  struct passwd *pwent;
  for (tail = user_alist; tail; tail = tail->next)
    if (tail->id.u == uid)
      return tail->name;
  pwent = getpwuid(uid);
  tail = (struct userid *)xmalloc(sizeof(struct userid));
  tail->id.u = uid;
  tail->name = pwent ? strdup (pwent->pw_name) : NULL;
  /* Add to the head of the list, so most recently used is first.  */
  tail->next = user_alist;
  user_alist = tail;
  return tail->name;
}

/* From the GNU fileutils package */
char *getgroup (gid_t gid) {
  struct userid *tail;
  struct group *grent;
  for (tail = group_alist; tail; tail = tail->next)
    if (tail->id.g == gid)
      return tail->name;
  grent = getgrgid (gid);
  tail = (struct userid *)xmalloc(sizeof (struct userid));
  tail->id.g = gid;
  tail->name = grent ? strdup (grent->gr_name) : NULL;
  /* Add to the head of the list, so most recently used is first.  */
  tail->next = group_alist;
  group_alist = tail;
  return tail->name;
}

FILE *display_directory(const char *filename) {
  DIR *dir = opendir(filename);
  FILE *file = tmpfile();
  struct dirent **entry;
  struct stat st;
  char mode[11], stime[13];
  int n, cnt;
  mode[10] = '\0';
  chdir(filename);
  fprintf(file,"Mode           User    Group       Size Modified     Name\n");
  fprintf(file,
          "-----------------------------------------------------------\n");
  n = scandir(filename, &entry, (void*)one, alphasort);
  for (cnt = 0; cnt < n; ++cnt) { /* GNU libc info files */
    stat(entry[cnt]->d_name, &st);
    mode_string(st.st_mode, mode);
    strftime(stime, 15, "%b %e %H:%M", localtime(&st.st_mtime));
    fprintf(file, "%s %8s %8s %10d %s %s\n", mode, getuser(st.st_uid),
            getgroup(st.st_gid), (int)st.st_size, stime, entry[cnt]->d_name);
  }
  closedir(dir);
  return file;
}

void parse_stdin(FileNode *n) {
  char line[1024], section[2], man[64];
  n->is_file = 0;
  n->file = tmpfile();
  waddstr(text, "Please wait... awaiting input.");
  wrefresh(text);
  fgets(line, 1024, stdin);
  if (sscanf(line, "%63[^( ](%1[0-9]) %*s", man, section) == 2) {
    int i;
    n->filename = xmalloc(sizeof(char) * 128);
    for (i = 0; i < strlen(man); i++)
      man[i] = tolower(man[i]);
    snprintf(n->filename, 128, "%s(%s)", man, section);
  } else n->filename = strdup("Input Stream");
  fputs(line, n->file);
  while (fgets(line, 1024, stdin)) fputs(line, n->file);
  freopen("/dev/tty", "r", stdin);
}

FileNode *load_file(const char *filename) {
  FileNode *node = new_file_node();
  struct stat st;
  char *newfilename = NULL;
  node->filename = xmalloc(sizeof(char) * (PATH_MAX + 1));
  if (strcmp(filename, "-")) { /* Not stdin... */
    if (strstr(filename, "~/") == filename) {
      newfilename = xmalloc(sizeof(char) * (strlen(filename) +
                                            strlen(getenv("HOME"))));
      sprintf(newfilename, "%s/%s", getenv("HOME"), filename + 2);
    } else newfilename = strdup(filename);
    if (realpath(newfilename, node->filename) == NULL) {
      delete_file_node(node);
      free(newfilename);
      return NULL;
    }
    node->filename = realloc(node->filename,
                             sizeof(char) * (strlen(node->filename) + 1));
    stat(node->filename, &st);
    if (S_ISDIR(st.st_mode)) node->file = display_directory(node->filename);
    else node->file = fopen(node->filename, "r");

    if (node->file == NULL) {
      free(node->filename);
      free(node);
      free(newfilename);
      return NULL;
    }
  } else parse_stdin(node);

  rewind(node->file);
  free(newfilename);
  return node;
}

void find_line_markers(FileNode *node) {
  unsigned int length = 100, i;
  node->lines = node->line = 0;
  if (node->markers != NULL) free(node->markers);
  node->markers = xmalloc(sizeof(int) * length);
  node->markers[0] = 0;
  rewind(node->file);

  while (!feof(node->file)) {
    for (i = 0; i < COLS; i++) {
      switch (fgetc(node->file)) {
      case '\b': i -= 2; break;
      case '\n': i = COLS - 1; break;
      case '\t': i = (((i / 8) + 1) * 8) - 1; break;
      case '\0': case '\a': case 255: i--; break;
      }
    }
    node->lines++;
    if (node->lines == length) {
      length *= 2;
      node->markers = realloc(node->markers, sizeof(int) * length);
    }
    node->markers[node->lines] = ftell(node->file);
  }
  node->markers = realloc(node->markers, sizeof(int) * (node->lines + 1));
}

void pause_file(FileNode *node) {
  node->paused = 1;
  update_status_bar(node);
  nodelay(text, FALSE);
}

void unpause_file(FileNode *node) {
  node->paused = 0;
  update_status_bar(node);
  nodelay(text, TRUE);
}

void toggle_pause(FileNode *node) {
  node->paused == 1 ? unpause_file(node) : pause_file(node);
}

int seek_to_line_and_display(FileNode *node, int line) {
  int i;

  if (line > (node->lines - (LINES - 3))) {
    line = (node->lines - (LINES - 3));
    if (node->jump > 0) pause_file(node);
    if (node->lines > (LINES - 3)) alert();
  }

  if (line < 0) {
    line = 0;
    if (node->lines > (LINES - 3)) alert();
    wclear(text);
  }

  fseek(node->file, node->markers[line], SEEK_SET);

  for (i = 0; i < LINES - 4 && line + i + 1 < node->lines; i++) {
    unsigned char l[node->markers[line + i + 1] -
                    node->markers[line + i] + 1];
    char *dos_style_newline = NULL;
    regmatch_t matched[5];

    fgets(l, (int)(node->markers[line + i + 1] -
                   node->markers[line + i]) + 1, node->file);

    dos_style_newline = strstr(l, "\r\n");
    if (dos_style_newline != NULL) dos_style_newline[0] = ' ';

    matched[0].rm_eo = matched[0].rm_so = -1;
    if (index(l, '\b') != NULL ||
        (node->match != NULL && !regexec(node->match, l, 5, matched, 0))) {
      int j;
      for (j = 0; j < strlen(l); j++) {
        if (j >= matched[0].rm_so && j < matched[0].rm_eo)
          wattron(text, A_STANDOUT);

	if (l[j] == '\b') {
	  waddch(text, '\b');
	  if (l[j - 1] == l[j + 1]) wattron(text, A_BOLD);
	  else if (l[j - 1] == '_') wattron(text, A_UNDERLINE);
	  waddch(text, l[++j]);
	} else waddch(text, l[j]);
        wattroff(text, A_BOLD | A_UNDERLINE | A_STANDOUT);
      }
    } else waddstr(text, l);	/* Just do it all at once, if possible */
  }
  wrefresh(text);
  return line;
}

void update_status_bar(FileNode *node) {
  char *message = xmalloc(sizeof(char) * (COLS + 1)), *paused = "";

  mvwhline(status, 0, 0, ' ', COLS);
  mvwhline(status, 1, 0, ' ', COLS);

  if (node->message == NULL) {
    mvwaddstr(status, 0, 0, "File: ");
    if (strlen(node->filename) > COLS - 6) {
      waddstr(status, "[...]");
      waddstr(status, node->filename + (strlen(node->filename) - COLS + 11));
    } else waddstr(status, node->filename);
  } else mvwaddstr(status, 0, 0, node->message);

  if (node->paused) paused = "Paused. ";
  snprintf(message, COLS + 1, "%sDelay: %.1f, Jump: %d. %d/%d, %.1f%%",
	   paused, ((float) node->delay) / 100000, node->jump, node->line,
	   node->lines, (float)(node->line) / (float)(node->lines) * 100.0);
  mvwaddstr(status, 1, COLS - strlen(message), message);
  wrefresh(status);
  free(message);
}

char *input_string(const char *message) {
  char *value = xmalloc(sizeof(char) * 128);
  mvwhline(status, 0, 0, ' ', COLS);
  echo();
  curs_set(1);
  waddstr(status, message);
  wgetnstr(status, value, 128);
  noecho();
  curs_set(0);
  return value;
}

char input_char(const char *message) {
  char c = ERR;
  mvwhline(status, 0, 0, ' ', COLS);
  waddstr(status, message);
  c = wgetch(status);
  /* If this isn't done, weird things happen, e.g. pressing right arrow
     generates ^[[C, which then executes the C command and (to the user's
     surprise) clears all bookmarks. */
  nodelay(status, TRUE);
  while (wgetch(status) != ERR) {}
  nodelay(status, FALSE);
  return c;
}

float input_decimal(const char *message) {
  char *tmp = input_string(message);
  float value = atof(tmp);
  free(tmp);
  return value;
}

int input_integer(const char *message) {
  return (int)(input_decimal(message));
}

int get_bookmark(const char *name, FileNode *node) {
  char line[1024], file[PATH_MAX + 1], markname[128];
  int lineno = 0;
  FILE *bf = fopen(bookmark_file, "r");
  if (bf == NULL) return -1;

  while (fgets(line, 1024, bf)) {
    if (sscanf(line, "%[^\t]\t%d\t%127s\n", file, &lineno, markname) == 3) {
      if (strcmp(file, node->filename) == 0 && strcmp(markname, name) == 0) {
	fclose(bf);
	return lineno;
      }
    }
  }
  fclose(bf);
  return -1;
}

void delete_bookmark(const char *name, FileNode *node) {
  char line[1024], tmp[strlen(getenv("HOME")) + 14];
  FILE *bf = fopen(bookmark_file, "r"), *tmpf = NULL;
  sprintf(tmp, "%s/.reed_XXXXXX", getenv("HOME"));
  tmpf = fdopen(mkstemp(tmp), "a");
  if (tmpf == NULL || bf == NULL) {
    if (tmpf != NULL) fclose(tmpf);
    if (bf != NULL) fclose(bf);
    return;
  }
  while (fgets(line, 1024, bf)) {
    char file[PATH_MAX + 1], markname[128];
    if (sscanf(line, "%[^\t]\t%*[0-9]\t%127s\n", file, markname) == 2 &&
	(strcmp(markname, name) || strcmp(file, node->filename)))
      fputs(line, tmpf);
  }
  fclose(bf); fclose(tmpf);
  unlink(bookmark_file);
  rename(tmp, bookmark_file);
}

void set_bookmark(const char *name, FileNode *node) {
  FILE *bf;
  int i;
  if (get_bookmark(name, node) != -1) delete_bookmark(name, node);
  bf = fopen(bookmark_file, "a");
  if (bf == NULL) return;
  fprintf(bf, "%s", node->filename);
  for (i = 50 - strlen(node->filename); i > 0; i -= 8) fputc('\t', bf);
  fputc('\t', bf);
  fprintf(bf, "%d\t%s\n", node->line, name);
  fclose(bf);
}

void clear_bookmarks(FileNode *node) {
  char line[PATH_MAX + 140], tmp[strlen(getenv("HOME")) + 14];
  FILE *bf = fopen(bookmark_file, "r"), *tmpf;
  sprintf(tmp, "%s/.reed_XXXXXX", getenv("HOME"));
  tmpf = fdopen(mkstemp(tmp), "a");
  if (bf) {
    while (fgets(line, PATH_MAX + 140, bf)) {
      char *original = strdup(line);
      char *file = strtok(line, "\t");
      if (strcmp(file, node->filename)) fprintf(tmpf, original);
      free(original);
    }
    fclose(bf);
  }
  unlink(bookmark_file);
  rename(tmp, bookmark_file);
  fclose(tmpf);
}

void set_char_bookmark(char c, FileNode *node) {
  char s[2] = " "; s[0] = c;
  set_bookmark(s, node);
}

int get_char_bookmark(char c, FileNode *node) {
  char s[2] = " "; s[0] = c;
  return get_bookmark(s, node);
}

regex_t *make_regexp(const char *s) {
  regex_t *match = xmalloc(sizeof(regex_t));
  if (regcomp(match, s, 0))
    return NULL;
  return match;
}

int search_for(regex_t *match, FileNode *node, int up_down) {
  int j, to, found = -1;
  if (match == NULL)
    return node->line;

  if (up_down == 1) {
    fseek(node->file, node->markers[node->line + 1], SEEK_SET);
    j = node->line + 1;
    to = node->lines;
  } else {
    fseek(node->file, node->markers[0], SEEK_SET);
    j = 0;
    to = node->line;
  }
  for (; j < to; j++) {
    int num = node->markers[j+1] - node->markers[j];
    char line[num + 1];
    if (fgets(line, num+1, node->file)) {
      if (!(regexec(match, line, 0, NULL, 0))) {
	found = j;
	if (up_down == 1) break;
      }
    }
    else break;
  }
  return found;
}

void new_search(FileNode *node, int up_down) {
  char *s = input_string("Search for: ");
  int new_line;

  if (!strcmp(s, "")) {
    free(s);
    return;
  } else if (node->match != NULL) {
    free_regexp(node->match);
  }

  if ((node->match = make_regexp(s)) == NULL) {
    alert();
    node->message = "Error: Bad regular expression.";
    free(s);
    return;
  }
  node->line -= up_down;
  new_line = search_for(node->match, node, up_down);
  if (new_line >= 0) {
    node->line = seek_to_line_and_display(node, new_line);
  } else if (new_line == -1) {
    node->line += up_down;
    node->message = "The search pattern was not found.";
    node->line = seek_to_line_and_display(node, node->line);
    free_regexp(node->match);
  }
  free(s);
  pause_file(node);
  return;
}

void spawn_editor(FileNode *node) {
  char *editor, *cline;
  if (getenv("VISUAL")) { editor = getenv("VISUAL"); }
  else if (getenv("EDITOR")) { editor = getenv("EDITOR"); }
  else { editor = "/bin/ed"; }
  cline = xmalloc(sizeof(char) *
		  (strlen(editor) + strlen(node->filename) + 16));
  sprintf(cline, "%s +%d \"%s\"", editor, node->line, node->filename);
  system(cline);
  free(cline);
}

void next_bracket(const char *left, const char *right, FileNode *node,
                  int direction) {
  regex_t *left_match = NULL, *right_match = NULL;
  int left_pos, right_pos, found = -1, old_pos = node->line;
  left_match = make_regexp(left);
  right_match = make_regexp(right);
  if (direction == 1) {
    node->line--;
    left_pos = search_for(left_match, node, 1);
    if (left_pos == -1) {
      node->message = "Error: The start character was not found.";
      free_regexp(left_match);
      free_regexp(right_match);
      return;
    }
    right_pos = left_pos;
  } else {
    node->line++;
    right_pos = search_for(right_match, node, -1);
    if (right_pos == -1) {
      node->message = "Error: The start parenthesis was not found.";
      node->line = old_pos;
    }
    left_pos = right_pos;
  }

  while (found == -1) {
      node->line = left_pos;
      left_pos = search_for(left_match, node, direction);
      node->line = right_pos;
      right_pos = search_for(right_match, node, direction);

      if ((right_pos == -1 && direction == 1) ||
          (left_pos == -1 && direction == -1)) {
        node->message = "No matching parenthesis was found.";
        found = old_pos;
      }
      else if (((right_pos < left_pos || left_pos == -1) && direction == 1) ||
               ((left_pos > right_pos || right_pos == -1) && direction == -1)) {
        if (direction == 1) found = right_pos;
        if (direction == -1) found = left_pos;
      }
    }
    free_regexp(right_match);
    free_regexp(left_match);
    node->line = seek_to_line_and_display(node, found);
}


int set_active_file_node(FileNode *node) {
  unsigned long count = 0;
  int c;
  node->line = seek_to_line_and_display(node, node->line);
  node->paused ? pause_file(node) : unpause_file(node);
  update_status_bar(node);
  while ((c = wgetch(text)) != 'q') {
    node->message = NULL;

    switch (c) {
    case ERR: break;
    case KEY_UP:
    case 'k':
      node->line = seek_to_line_and_display(node, node->line - 1);
      break;
      
    case KEY_DOWN:
    case 'e':
    case 'j':
      node->line = seek_to_line_and_display(node, node->line + 1);
      break;
      
    case KEY_NPAGE:
    case ' ':
    case 'z':
      node->line =
	seek_to_line_and_display(node, node->line + (LINES - 4));
      break;
      
    case KEY_PPAGE:
    case 'w':
    case 'b':
      node->line =
	seek_to_line_and_display(node, node->line - (LINES - 4));
      break;
      
    case 'u':
      node->line =
	seek_to_line_and_display(node, node->line - (LINES - 4) / 2);
      break;
      
    case 'd':
      node->line =
	seek_to_line_and_display(node, node->line + (LINES - 4) / 2);
      break;
      
    case 'p':
      toggle_pause(node);
      break;
      
    case '\r':
    case '\n':
      if (node->paused)
        node->line = seek_to_line_and_display(node, node->line + 1);
      else count = node->delay;
      break;
      
    case '-':
    case '_':
      if (node->delay > 100000)
	node->delay -= 20000;
      else
	node->delay /= 2;
      break;
      
    case '+':
    case '=':
      node->delay += 20000;
      break;
      
    case '>':
    case '.':
    case 'G':
    case KEY_END:
      node->line = seek_to_line_and_display(node, node->lines - (LINES - 3));
	    break;
	    
    case '<':
    case ',':
    case 'g':
    case KEY_HOME:
      node->line = seek_to_line_and_display(node, 0);
      break;
	    
    case 's':
      node->delay =
	input_decimal("Enter the time to wait between lines: ") * 100000;
      break;

    case 'J':
      node->jump =
        input_integer("Enter the number of lines to skip at a time: ");
      break;

    case 'm':{
      char *s = input_string("Go to bookmark: ");
      int new_line = 0;
      if (s != NULL) new_line = get_bookmark(s, node);
      if (new_line == -1) {
	alert();
	node->message = "The bookmark was not found.";
      } else node->line = seek_to_line_and_display(node, new_line);
      pause_file(node);
      free(s);
      break;
    }
      
    case 'M':{
      char *s = input_string("Set bookmark: ");
      if (strlen(s)) set_bookmark(s, node);
      free(s);
      break;
    }
      
    case 'D':{
      char *s = input_string("Delete bookmark: ");
      delete_bookmark(s, node);
      free(s);
      break;
    }

    case 'C': clear_bookmarks(node); break;
    case 'B': return BOOKMARKS;

    case '\'':{
      int new_line = get_char_bookmark(input_char("Go to "), node);
      if (new_line == -1) node->message = "The bookmark was not found.";
      else node->line = seek_to_line_and_display(node, new_line);
      pause_file(node);
      break;
    }

    case '"':
      set_char_bookmark(input_char("Set "), node);
      break;

    case 'l':{
      int new_line = get_bookmark("default", node);
      if (new_line == -1) {
        alert();
        node->message = "The bookmark was not found.";
      } else node->line = seek_to_line_and_display(node, new_line);
      pause_file(node);
      break;
    }

    case 'L':
      set_bookmark("default", node);
      break;

    case '/': new_search(node, 1); break;
    case 'n':{
      int new_line = search_for(node->match, node, 1);
      if (new_line != -1)
        node->line = seek_to_line_and_display(node, new_line);
      break;
    }
    case '?': new_search(node, -1); break;
    case 'N':{
      int new_line = search_for(node->match, node, -1);
      if (new_line != -1)
        node->line = seek_to_line_and_display(node, new_line);
      break;
    }

    case '(': next_bracket("(",")", node, 1); break;
    case '{': next_bracket("{","}", node, 1); break;
    case '[': next_bracket("\\[","\\]", node, 1); break;
    case ')': next_bracket("(",")", node, -1); break;
    case '}': next_bracket("{","}", node, -1); break;
    case ']': next_bracket("\\[","\\]", node, -1); break;

    case 'F':
      free_regexp(node->match);
      node->line = seek_to_line_and_display(node, node->line);
      break;

    case '!':{
      char *s = input_string("Run command: ");
      endwin();
      system(s);
      free(s);
      initialize_windows();
      node->line = seek_to_line_and_display(node, node->line);
      break;
    }

    case '%':
      node->line = seek_to_line_and_display(node,
        node->lines * (input_decimal("Go to percentage: ") / 100.0));
      break;

    case 'v':
      spawn_editor(node);
      initialize_windows();
      seek_to_line_and_display(node, node->line);
      break;

    case 'a':
      cue ^= 1;
      break;

    case 't':
      node->line = seek_to_line_and_display(node,
                                            input_integer("Go to line: "));
      break;

    case 'r':
      initialize_windows();
      seek_to_line_and_display(node, node->line);
      break;

    case 'R':
      free(node->filename);
      node->filename = input_string("Enter a new file name: ");
      node->is_file = 0;
      break;

    case 'H':
      return HELP;

    case ':':
      switch (input_char("n)ext, p)revious, d)elete, e)xamine, r)eload, or q)uit? ")) {
      case 'n':	return NEXT_FILE;
      case 'p':	return PREV_FILE;
      case 'd':	return DELETE_FILE;
      case 'e':	return LOAD_FILE;
      case 'r': return RELOAD_FILE;
      case 'q':	return QUIT;
      }
      break;

    default:
      node->message = "Unknown command. Try 'H' for help.";
    }

    update_status_bar(node);
    if (!(node->paused) && node->jump != 0) {
      int num_jumps = abs(node->jump);
      count += 20000;
      usleep(10000); /* usleep(1) below takes 10ms on GNU/Linux i386 */
      if (count > (node->delay * num_jumps)) {
	int direction = num_jumps / node->jump, i = 0;
	count = 0;
	for (i = 0; i < num_jumps &&
	     ((node->line < node->lines - (LINES - 3) && direction > 0) ||
	      (node->line > 0 && direction < 0));
	     i++) {
	  node->line = seek_to_line_and_display(node, node->line + direction);
	  update_status_bar(node);
	  usleep(1);
	}
	if ((node->jump < 0 && node->line == 0) ||
	    (node->jump > 0 && node->line == node->lines - (LINES - 3)))
	  pause_file(node);
      }
    }
  }
  return QUIT;
}

FileNode *load_new_node(FileNode *old_node, const char *filename) {
  FileNode *new_node = load_file(filename);
  if (new_node == NULL) {
    alert();
    old_node->message = "There was an error loading the file.";
    return old_node;
  } else {
    new_node->n = old_node->n;
    new_node->p = old_node;
    old_node->n = new_node;
    return new_node;
  }
}

void parse_rc_file() {
  char rcfile[strlen(getenv("HOME")) + 9];
  FILE *rc;
  sprintf(rcfile, "%s/.reedrc", getenv("HOME"));
  if ((rc = fopen(rcfile, "r")) != NULL) {
    char line[512];
    while (fgets(line, 512, rc)) {
      char key[10], value[10];
      if (sscanf(line, "%10s %10s\n", key, value) == 2) {
        if (!strcasecmp(key, "delay")) initial_delay = (atof(value) * 100000);
        else if (!strcasecmp(key, "jump")) initial_jump = atoi(value);
        else if (!strcasecmp(key, "beep") && !strcasecmp(value, "off"))
          cue = 0;
       else if (!strcasecmp(key, "paused") && !strcasecmp(value, "off"))
          initial_pause = 0;
      }
    }
    fclose(rc);
  }
}

void print_help() {
  fprintf(stderr, "Reed %s - An autoscrolling text pager\n", VERSION);
  fprintf(stderr, "Proper usage: reed [options] [filenames]\n");
  fprintf(stderr, "Options (overriding ~/.reedrc):\n");
  fprintf(stderr, " -d [num]\tScrolling delay in approximately tenths of a second.\n");
  fprintf(stderr, " -j [num]\tNumber of lines to jump when scrolling.\n");
  fprintf(stderr, " -u/-p\t\tStarted unpaused or paused, respectively.\n");
  fprintf(stderr, " -b/-q\t\tTurn audio cues on or off, respectively.\n");
  fprintf(stderr, " -v\t\tDisplay version and copyright information.\n");
  fprintf(stderr, " -h\t\tDisplay this help text.\n");
  exit(0);
}

void print_version() {
  fprintf(stderr, "Reed %s - An autoscrolling text pager\n", VERSION);
  fprintf(stderr, "Copyright (c)2000-2002 Joe Wreschnig\n\n");
  fprintf(stderr, "This program is free software; you can redistribute it and/or modify\n"); 
  fprintf(stderr, "it under the terms of the GNU General Public License as published by\n"); 
  fprintf(stderr, "the Free Software Foundation; either version 2 of the License, or\n");
  fprintf(stderr, "(at your option) any later version.\n\n");
  fprintf(stderr, "This program is distributed in the hope that it will be useful,\n");
  fprintf(stderr, "but WITHOUT ANY WARRANTY; without even the implied warranty of\n");
  fprintf(stderr, "MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n");
  fprintf(stderr, "GNU General Public License for more details.\n");
  exit(0);
}

void initialize_windows(void) {
  if (status != NULL) delwin(status);
  if (text != NULL) delwin(text);
  status = newwin(2, COLS, 0, 0);
  text = newwin(LINES - 3, COLS, 3, 0);
  mvhline(2, 0, '-', COLS); refresh();
  scrollok(text, TRUE); noecho(); cbreak(); keypad(text, TRUE); curs_set(0);
  wclear(text); wclear(status);
}

int main(int argc, char **argv) {
  char *f = NULL, opt;
  int action = 0, i = 0;
  struct stat st;
  FileNode *active = NULL;

  if (argc > 1 && !strcmp(argv[1], "--help")) print_help();

  bookmark_file = xmalloc(sizeof(char) * (strlen(getenv("HOME")) + 17));
  sprintf(bookmark_file, "%s/.reed_bookmarks", getenv("HOME"));

  parse_rc_file();

  while ((opt = getopt(argc, argv, "bqhvd:j:puf:")) != -1) {
    switch (opt) {
    case 'h': print_help();
    case 'v': print_version();
    case 'd': initial_delay = atof(optarg) * 100000; break;
    case 'j': initial_jump = atoi(optarg); break;
    case 'p': initial_pause = 1; break;
    case 'u': initial_pause = 0; break;
    case 'b': cue = 1; break;
    case 'q': cue = 0; break;
    case 'f': f = optarg;
    }
  }

  initscr();
  
  for (i = optind; i < argc; i++) {
    int valid = stat(argv[i], &st);
    if (valid == -1 && strcmp(argv[i], "-")) {
      fprintf(stderr, "reed: warning: %s: Invalid filename. Ignoring it...\n",
	      argv[i]);
    } else {
      if (active == NULL) active = load_file(argv[i]);
      else {
	FileNode *new_node = load_file(argv[i]);
	if (new_node != NULL) {
	  active->n = new_node;
	  new_node->p = active;
	  active = new_node;
	}
      }
      if (f != NULL) {
        free(active->filename);
        active->filename = strdup(f);
        active->is_file = 0;
      }
    }
  }
  
  if (active == NULL) {
    if (isatty(fileno(stdin)) == 0) {
      active = load_file("-");
    } else {
      endwin();
      fprintf(stderr, "reed: error: No valid filenames were found.\n");
      exit(1);
    }
  }

  initialize_windows();

  while (active->p != NULL) {
    active = active->p;
  }

  waddstr(text, "Please wait... loading file.\n");
  wrefresh(text);
  find_line_markers(active);
  
  while ((action = set_active_file_node(active)) != QUIT) {
    wclear(status);
    wclear(text);
    switch (action) {
    case NEXT_FILE:
      active = next_file_node(active);
      break;
    case PREV_FILE:
      active = prev_file_node(active);
      break;
    case DELETE_FILE:
      if (active->p) {
	active = active->p;
	delete_file_node(active->n);
      } else if (active->n) {
	active = active->n;
	delete_file_node(active->p);
      } else {
	delete_file_node(active);
	active = NULL;
      }
      break;
    case RELOAD_FILE:
      initialize_windows();
      if (active->is_file) {
       active = load_new_node(active, active->filename);
       delete_file_node(active->p);
      }
      find_line_markers(active);
      break;
    case LOAD_FILE:
      f = input_string("Enter a filename to load: ");
      active = load_new_node(active, f);
      free(f);
      break;
    case BOOKMARKS:
      if (strcmp(active->filename, "Your Bookmarks")) {
        active = load_new_node(active, bookmark_file);
        if (!strcmp(active->filename, bookmark_file)) {
	  free(active->filename);
	  active->filename = strdup("Your Bookmarks");
	  active->is_file = 0;
        }
      }
      break;
    case HELP:
      if (strcmp(active->filename, "Help using Reed")) {
        FileNode *new_node = load_new_node(active, "/usr/local/share/reed/help");
        if (new_node == active)
          new_node = load_new_node(active, "/usr/share/reed/help");
        if (new_node != active) {
          active->message = NULL;
          active = new_node;
	  free(active->filename);
	  active->filename = strdup("Help using Reed");
	  active->is_file = 0;
        }
      }
      break;
    }
    if (active == NULL) break;
    else if (active->markers == NULL) {
      waddstr(text, "Please wait... loading file.\n");
      wrefresh(text);
      find_line_markers(active);
    }
  }
  endwin();
  exit(0);
}
